#ifndef DIN
#define DIN

#include "curve_editor.h"
#include "strength_editor.h"
#include "listeners.h"
#include "range.h"
#include "resonator.h"
#include "modder.h"
#include "volume.h"
#include "ui.h"
#include "audio.h"
#include "play.h"
#include "oscilloscope.h"
#include "curve_library.h"
#include "beat2value.h"
#include "help.h"
#include "phrasor.h"
#include "input.h"
#include "scale_info.h"
#include "command.h"
#include "compressor.h"
#include "globals.h"

using namespace std;

struct drone { // drone tone

  play player; // player for this drone

  float step; // determines frequency
  float vol; // actual volume

	int range_num; // range number
	float fpos; // range position 0 .. 1

	int ipos; // for drawing drone maker and picking
  int dv; // driver volume

};

struct din_info {

	string resonators_fname; // resonators file

	point<int> scroll; // x and y scroll speed
	int height; // din board height

	int delay; // is delay enabled?

	int compress; // is compressor enabled?

	// multicore vars
	int ncores;
	int resonators_per_core;
	int last_resonator;

	din_info ();
	~din_info ();

};

struct din;

struct octave_shift_handler : xhandler {
	din* d;
	float tonic;
	int dir;
	octave_shift_handler (din* dd) : d(dd), tonic(0), dir(1){}
	void operator() (solver& s, float& x, float& dx);
};

struct din : ui {

  // stereo volume of all resonators
	static const int MAX_VOLUMES = 1024;
  volume volumes [MAX_VOLUMES];

  // tone ranges
	static const int MAX_RANGES = 64;
	range ranges [MAX_RANGES];
  int num_ranges, last_range, current_range;
  range *firstr, *lastr;
	float step, delta, octave_position;

  void setup_ranges ();
	void setup_range_notes ();
	void update_range_notes ();
	void range_right_changed ();
	void range_left_changed ();
	bool load_ranges ();
	void set_range_size (int r, int sz);
	void save_ranges ();
	void find_current_range ();
	bool find_tone_and_volume ();
	void retune_note (const string& nn, float v);
	void retune_note ();
	void restore_note ();
	void restore_all_notes ();
	float get_note_value (const string& s);

  // mouse board
  int win_mousex, win_mousey; // mouse pos in win
  int prev_mousex, prev_mousey;
  int delta_mousex, delta_mousey;
  void calc_win_mouse ();

  int dv; // driver volume ie mouse height on board
  note* K; // current key data
	box<int> win;
  void scroll (int dx, int dy);
	void set_tonic (float sa, int warp = 0);
	void mouse2tonic ();
	float get_tonic ();


	//
	// drones
	//
  int num_drones;
  vector<drone> drones;
  float drone_master_volume;
  multi_curve drone_wave; // drone waveform
  solver drone_sol; // solver for drone waveform
  curve_editor droneed; // drone waveform editor
  drone_listener dronelis; // handles drone edits
  int current_drone;
	void load_drones ();
	void save_drones ();
	void sync_drones ();
	void add_drone ();
	void set_drone (int d);
	void pick_drone ();
	void remove_drone ();
	void mark_drones ();
	void update_drone_tone ();
	void update_drone_ipos (int i, int j);
	void update_drone_master_volume (float delta);
	int snap_drones;

	inline void find_volume () {
		// cursor height on din keyboard.
		extern int BOTTOM, LAST_VOLUME;
		dv = win_mousey - BOTTOM;
		if (dv > LAST_VOLUME) dv = LAST_VOLUME;
	}

	// am, fm, gater, octave shift
	float am_depth, am_vol;
	float fm_depth, fm_step;
	beat2value fm, am;
  beat2value gatl, gatr;
  octave_shift_handler osh;
  int is_octave_shift;
  beat2value octave_shift;

	void change_depth (int i, float d);
	void fill ();

	// beat2value editors
	curve_editor gated;
	curve_library gatlib;
	beat2value_listener gatllis, gatrlis;

  curve_editor octed;
  curve_library octlib;
  beat2value_listener octlis;

	// am & fm modulation editors
	curve_editor moded;
	beat2value_listener fmlis, amlis;

  curve_library wavlib; // waveform library

  //
  // resonators
  //

  std::vector<play> players;
  std::vector<resonator> resonators;

	//resonator resonators[resonator::MAX_RESONATORS];
	//play players [resonator::MAX_RESONATORS];

  int num_resonators;
	int cur_resonator;
	void add_resonator ();
	void edit_resonator (int i, unsigned int what);
	void select_resonator (int sel);
	void disable_resonator (int i, bool dis = true);
	void disable_all_resonators ();
	void render_left_volume (int r);
	void render_right_volume (int r);
	void render_volumes ();
	void render_resonator_strengths (range& ran);
	void load_resonators (const string& fname);
	void save_resonators (int deldis = 0);
	void browse_resonator (int i);
	void only_enable_resonator (int i);
	void enable_all_resonators ();

  // resonator editors
  strength_editor stred; // resonator.strength
  curve_editor waved; // resonator.wave
  curve_editor chaned; // edits both resonator.left and resonator.right

  // listeners interested in resonator edits
  strength_listener strlis;
  wave_listener wavelis;
  left_listener leftlis;
  right_listener rightlis;

  scale_info scaleinfo;

	// musical key
	int key_id; // C = 0, C# = 1, D = 2, D# = 3, E = 4, F = 5, F# = 6, G = 7, G# = 8, A = 9, A# = 10, B = 11
	void set_key_id (int i);
	int get_key_id () { return key_id;}

	// phrasor - record/play/scratch a phrase
	phrasor phrasor0;
	int jogged;

  oscilloscope osc;

	// compressor
	//
	curve_editor comed;
	compressor coml, comr;

	din (cmdlist& cmdlst);
	~din ();

	bool handle_input ();

	void load_scale ();
	void save_scale ();

	void enter ();
	void bg (); // background execution
	void draw ();

	// din saved info
	din_info dinfo;
	void calc_multicore_vars ();

  // help
  help helptext;

	// range display notation
	string notation;
	void set_notation (const string& s);
	void notate_ranges ();


};

#endif
