#ifndef __SOLVER
#define __SOLVER

#include "crvpt.h"
#include "multi_curve.h"
#include <vector>

struct solver;

struct xhandler {
	virtual void operator() (solver& s, float& x, float& dx) = 0;
	virtual ~xhandler () {}
};

struct atmin: xhandler {
	inline void operator() (solver& s, float& x, float& dx);
};

struct atmax: xhandler {
	inline void operator() (solver& s, float& x, float& dx);
};


struct tomin: xhandler {
	inline void operator() (solver& s, float& x, float& dx);
};

struct loopmin : atmin {
	inline void operator() (solver& s, float& x, float& dx);
};

struct loopmax : tomin {
	inline void operator() (solver& s, float& x, float& dx);
};

struct pongmax: xhandler {
	inline void operator() (solver& s, float& x, float& dx);
};

struct pongmin: xhandler {
	inline void operator() (solver& s, float& x, float& dx);
};


extern atmin _atmin;
extern tomin _tomin;
extern loopmin _loopmin;
extern loopmax _loopmax;
extern pongmin _pongmin;
extern pongmax _pongmax;

struct solver {

  //
  // given x return y
  //
  // assumes x such that
  // x0 <= x1 <= x2 <= ...... <= xlast
  //

	// solved curve
  multi_curve* mcrv;

  int ncurvs, icurv, last_curv;
  int iseg;

  // line segment on curve expected to contain x
	float startx, starty;
	float endx, endy;
	float m, inf;
	float ycomp;

  // first and last points
  float firstx, firsty;
  float lastx, lasty;

	float result; // last result

  inline int inseg (float x, int c, int i) {

		if (c > -1 && c < ncurvs) {
			std::vector<curve>& curv = mcrv->curv;
			std::vector<crvpt>& vpts = curv[c].vpts;
			float leftx = vpts[i].x, rightx = vpts[i+1].x;
			if ((x >= leftx) && (x <= rightx)) {
				setseg (c, i);
				return 1;
			}
		}
		return 0;

  }

	inline void setseg (int c, int i) {

		std::vector<curve>& curv = mcrv->curv;
		std::vector<crvpt>& vpts = curv[c].vpts;

		crvpt& start = vpts[i];
		crvpt& end = vpts[i+1];

		startx = start.x;
		starty = start.y;
		endx = end.x;
		endy = end.y;
		m = start.m;
		inf = start.inf;
		if (inf == 1) ycomp = endy; else ycomp = starty - m * startx;

		icurv = c;
		iseg = i;

	}

  inline int lastseg (int c) {
		std::vector<curve>& curv = mcrv->curv;
		if (c > -1) {
			std::vector<crvpt>& vpts = curv[c].vpts;
			return vpts.size() - 2;
		}
		return 0;
	}

	inline int getsegs (int c) {
		if (c > -1 && c < ncurvs) {
			return (mcrv->curv[c].vpts.size() - 1);
		}
		return 0;
	}

	inline int searchleft (float x) {
		int lseg = iseg, lcurv = icurv;
		while (lcurv > -1) {
			if (--lseg < 0) {
				--lcurv;
				lseg = lastseg (lcurv);
			}
			if (inseg (x, lcurv, lseg)) return 1;
		}
		return 0;
	}

	inline int searchright (float x) {
		int rseg = iseg, rcurv = icurv;
		int rnsegs = getsegs (rcurv);
		while (rcurv < ncurvs) {
			if (++rseg < rnsegs); else {
				++rcurv;
				rseg = 0;
				rnsegs = getsegs (rcurv);
			}
			if (inseg (x, rcurv, rseg)) return 1;
		}
		return 0;
	}

	inline int findseg (float x) {

		int lseg = iseg, rseg = iseg;
		int lcurv = icurv, rcurv = icurv;
		int rnsegs = getsegs (rcurv);

		while ((lcurv > -1) || (rcurv < ncurvs)) {

			if (--lseg < 0) {
				--lcurv;
				lseg = lastseg (lcurv);
			}
			if (inseg (x, lcurv, lseg)) return 1;

			if (++rseg < rnsegs); else {
				++rcurv;
				rseg = 0;
				rnsegs = getsegs (rcurv);
			}
			if (inseg (x, rcurv, rseg)) return 1;

		}
		return 0;
	}

	solver ();
	solver (multi_curve* crv);

	void operator() (multi_curve* crv);
	float operator() (float x); // given x, solve for y
	void operator() (float& x, float& dx, int n, float* res, xhandler& xmin = _atmin, xhandler& xmax = _tomin, int dn = 1); // fast solver when x is incremented by dx n times, solution stored in res
	void operator() (float& x, float& dx, int n, float* mod, float* res, xhandler& xmin = _atmin, xhandler& xmax = _tomin, int dn = 1); // same as last solver but x is also modulated by mod.
	void operator() (float* ax, int n, int dx = 2, xhandler& xmin = _atmin, xhandler& xmax = _tomin); // given an array of x, store solution in same array.

	void init ();
	void update (); // used when multi curve has changed


};




#endif
