/**********************************************************************/
/*! \class RtMidi
    \brief An abstract base class for realtime MIDI input/output.

    This class implements some common functionality for the realtime
    MIDI input/output subclasses RtMidiIn and RtMidiOut.

    RtMidi WWW site: http://music.mcgill.ca/~gary/rtmidi/

    RtMidi: realtime MIDI i/o C++ classes
    Copyright (c) 2003-2010 Gary P. Scavone
    
		modified by S Jagannathan, July-August 2010.

*/
/**********************************************************************/

#include "RtMidi.h"
#include <sstream>

extern "C" void* alsaMidiHandler (void *ptr ) {
  
  RtMidiIn::RtMidiInData *data = static_cast<RtMidiIn::RtMidiInData *> (ptr);
  AlsaMidiData *apiData = data->apiData;

  int nbytes;
  bool issysex = false;
  std::vector<unsigned char> message;

  snd_seq_event_t *ev;  
  snd_midi_event_new( 0, &apiData->coder );
  snd_midi_event_init( apiData->coder );
  snd_midi_event_no_status( apiData->coder, 1 ); // suppress running status messages

  while ( data->doInput ) {

    if ( snd_seq_event_input_pending( apiData->seq, 1 )) {
			snd_seq_event_input( apiData->seq, &ev );
			if ( !issysex ) {
				message.clear();
				apiData->bufferSize = 32;
			}
			nbytes = snd_midi_event_decode( apiData->coder, apiData->buffer, apiData->bufferSize, ev);
			if (nbytes) {
					for (int i = 0; i < nbytes; ++i) message.push_back (apiData->buffer[i]);
					issysex = ( ( ev->type == SND_SEQ_EVENT_SYSEX ) && ( message.back() != 0xF7 ) );
			}
			data->queue.push( message );
			snd_seq_free_event(ev);
		}
		
  }
  snd_midi_event_free( apiData->coder );
  apiData->coder = 0;
  return 0;
}


RtMidi :: RtMidi(): connected (false), portnum (-1) {}

void RtMidi :: error( RtError::Type type )
{
  if (type == RtError::WARNING) {
    std::cerr << '\n' << errorString << "\n\n";
  }
  else {
    std::cerr << '\n' << errorString << "\n\n";
    throw RtError( errorString, type );
  }
}

//*********************************************************************//
//  Common RtMidiIn Definitions
//*********************************************************************//

RtMidiIn :: RtMidiIn ( const std::string clientName ) : RtMidi() {
  this->initialize( clientName );
}

void RtMidiIn :: setQueueSizeLimit( unsigned int queueSize ) {
  inputData.queueLimit = queueSize;
}

void RtMidiIn :: getMessage( std::vector<unsigned char> *message ) {

  message->clear();
  if ( inputData.queue.size() == 0 ) return;
  // Copy queued message to the vector pointer argument and then "pop" it.
  std::vector<unsigned char> *bytes = &inputData.queue.front();
  message->assign( bytes->begin(), bytes->end() );
  inputData.queue.pop();
}

//*********************************************************************//
//  API: LINUX ALSA SEQUENCER
//*********************************************************************//

// API information found at:
//   - http://www.alsa-project.org/documentation.php#Library

// The ALSA Sequencer API is based on the use of a callback function for
// MIDI input.
//
// Thanks to Pedro Lopez-Cabanillas for help with the ALSA sequencer
// time stamps and other assorted fixes!!!

// If you don't need timestamping for incoming MIDI events, define the
// preprocessor definition AVOID_TIMESTAMPING to save resources
// associated with the ALSA sequencer queues.

#include <pthread.h>
#include <sys/time.h>


#define PORT_TYPE( pinfo, bits ) ((snd_seq_port_info_get_capability(pinfo) & (bits)) == (bits))


void RtMidiIn :: initialize( const std::string& clientName ) {
  
  // Set up the ALSA sequencer client.
  snd_seq_t *seq;
  int result = snd_seq_open(&seq, "default", SND_SEQ_OPEN_DUPLEX, SND_SEQ_NONBLOCK);
  if ( result < 0 ) {
    errorString = "RtMidiIn::initialize: error creating ALSA sequencer input client object.";
    error ( RtError::DRIVER_ERROR );
	}

  // Set client name.
  snd_seq_set_client_name( seq, clientName.c_str() );

  // Save our api-specific connection information.
  apiData.seq = seq;
	apiData.vport = -1;
  inputData.apiData = &apiData;

}

// This function is used to count or get the pinfo structure for a given port number.
unsigned int portInfo( snd_seq_t *seq, snd_seq_port_info_t *pinfo, unsigned int type, int portNumber )
{
	snd_seq_client_info_t *cinfo;
  int client;
  int count = 0;
	snd_seq_client_info_alloca( &cinfo );

	snd_seq_client_info_set_client( cinfo, -1 );
	while ( snd_seq_query_next_client( seq, cinfo ) >= 0 ) {
    client = snd_seq_client_info_get_client( cinfo );
    if ( client == 0 ) continue;
		// Reset query info
		snd_seq_port_info_set_client( pinfo, client );
		snd_seq_port_info_set_port( pinfo, -1 );
		while ( snd_seq_query_next_port( seq, pinfo ) >= 0 ) {
      unsigned int atyp = snd_seq_port_info_get_type( pinfo );
      if ( ( atyp & SND_SEQ_PORT_TYPE_MIDI_GENERIC ) == 0 ) continue;
      unsigned int caps = snd_seq_port_info_get_capability( pinfo );
      if ( ( caps & type ) != type ) continue;
      if ( count == portNumber ) return 1;
      ++count;
		}
	}
  // If a negative portNumber was used, return the port count.
  if ( portNumber < 0 ) return count;
  return 0;
}

int RtMidiIn :: openPort( unsigned int portNumber, const std::string portName ) {

  if ( connected ) {
		std::cout << "port " << portNumber << " used by " << portName << std::endl;
		return -1;
  }

  unsigned int nSrc = this->getPortCount();
  if (nSrc < 1) {
		std::cout << "--- no MIDI input devices found. ---" << std::endl;
		return -2;
  }

	snd_seq_port_info_t *pinfo;
	snd_seq_port_info_alloca( &pinfo );
  std::ostringstream ost;
  AlsaMidiData *data = static_cast<AlsaMidiData *> (&apiData);
  if ( portInfo( data->seq, pinfo, SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ, (int) portNumber ) == 0 ) {
    std::cout << " --- port number: " << portNumber << " is invalid ---" << std::endl;
    return -3;
  }

  snd_seq_addr_t sender, receiver;
  sender.client = snd_seq_port_info_get_client( pinfo );
  sender.port = snd_seq_port_info_get_port( pinfo );
  receiver.client = snd_seq_client_id( data->seq );
  if ( data->vport < 0 ) {
    snd_seq_port_info_set_client( pinfo, 0 );
    snd_seq_port_info_set_port( pinfo, 0 );
    snd_seq_port_info_set_capability( pinfo,
                                      SND_SEQ_PORT_CAP_WRITE |
                                      SND_SEQ_PORT_CAP_SUBS_WRITE );
    snd_seq_port_info_set_type( pinfo,
                                SND_SEQ_PORT_TYPE_MIDI_GENERIC |
                                SND_SEQ_PORT_TYPE_APPLICATION );
    snd_seq_port_info_set_midi_channels(pinfo, 16);
    snd_seq_port_info_set_name(pinfo,  portName.c_str() );
    data->vport = snd_seq_create_port(data->seq, pinfo);
  
    if ( data->vport < 0 ) {
      std::cout << "ALSA error: creating input port: " << portNumber << " with name: " << portName << std::endl;
      return -4;
    }
  }

  receiver.port = data->vport;

  // Make subscription
  snd_seq_port_subscribe_malloc( &data->subscription );
  snd_seq_port_subscribe_set_sender(data->subscription, &sender);
  snd_seq_port_subscribe_set_dest(data->subscription, &receiver);
  if ( snd_seq_subscribe_port(data->seq, data->subscription) ) {
    std::cout << "ALSA error: cant subscribe to port " << portNumber << ' ' << portName << std::endl;
    return -5;
  }

  if ( inputData.doInput == false ) {
    // Start the input queue
    // Start our MIDI input thread.
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    pthread_attr_setschedpolicy(&attr, SCHED_OTHER);
    inputData.doInput = true;
    int err = pthread_create(&data->thread, &attr, alsaMidiHandler, &inputData);
    pthread_attr_destroy(&attr);
    if (err) {
      snd_seq_unsubscribe_port( data->seq, data->subscription );
      snd_seq_port_subscribe_free( data->subscription );
      inputData.doInput = false;
      std::cout << "ALSA error: cant start MIDI input thread" << std::endl;
      return -6;
    }
  }
  
  connected = true;
  portnum = portNumber;
  
  return 1;
}

void RtMidiIn :: closePort( void ) {
  
  if ( connected ) {
    
    AlsaMidiData *data = static_cast<AlsaMidiData *> (&apiData);
    snd_seq_unsubscribe_port( data->seq, data->subscription );
    snd_seq_port_subscribe_free( data->subscription );
    
    connected = false;
    portnum = -1;
    portname = "";
  
  }
  
}

RtMidiIn :: ~RtMidiIn() {
  
  // Close a connection if it exists.
  closePort();

  // Shutdown the input thread.
  AlsaMidiData *data = static_cast<AlsaMidiData *> (&apiData);
  if ( inputData.doInput ) {
    inputData.doInput = false;
    pthread_join( data->thread, NULL );
  }

  // Cleanup.
  if ( data->vport >= 0 ) snd_seq_delete_port( data->seq, data->vport );
  snd_seq_close( data->seq );

}

unsigned int RtMidiIn :: getPortCount() {
	
	snd_seq_port_info_t *pinfo;
	snd_seq_port_info_alloca( &pinfo );

  AlsaMidiData *data = static_cast<AlsaMidiData *> (&apiData);
  return portInfo( data->seq, pinfo, SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ, -1 );
}

std::string RtMidiIn :: getPortName( unsigned int portNumber ) {
  
  snd_seq_client_info_t *cinfo;
  snd_seq_port_info_t *pinfo;
  snd_seq_client_info_alloca( &cinfo );
  snd_seq_port_info_alloca( &pinfo );

  AlsaMidiData *data = static_cast<AlsaMidiData *> (&apiData);
  if ( portInfo( data->seq, pinfo, SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ, (int) portNumber ) ) {
    int cnum = snd_seq_port_info_get_client( pinfo );
    snd_seq_get_any_client_info( data->seq, cnum, cinfo );
    std::string pname;
    pname += snd_seq_client_info_get_name( cinfo );
    return pname;
  } 
  return 0;
}

