#include "console.h"
#include "font.h"
#include "input.h"
#include "command.h"
#include "tcl_interp.h"
#include "tokenizer.h"
#include "globals.h"

extern cmdlist cmdlst;

#define buf cmdlst.buf

extern string norm_us, shift_us;
string console::charset[] = {
	"abcdefghijklmnopqrstuvwxyz0123456789 .=-/;\\,[]'`",
	"ABCDEFGHIJKLMNOPQRSTUVWXYZ)!@#$%^&*( >+_?:|<{}\"~",
};

extern tcl_interp* p_interp;

decoration::decoration (float rr, float gg, float bb, int w) {
  r = rr;
  g = gg;
  b = bb;
  lw = w;
}

const float cc = 0.6;
const decoration console::light = decoration ();
const decoration console::yellow = decoration (1, 1, cc);
const decoration console::green = decoration (cc, 1, cc);
const decoration console::red = decoration (1, cc, cc);
const decoration console::cyan = decoration (cc, 1, 1);
const decoration console::bold = decoration (0, 0, 0, 2);

void console::push_back (const mesg& ln) {

  lines.push_back (ln);
  ++nlines;
  last ();

}

void console::up (int i) {
	startl = max (0, startl - i);
}

void console::down (int i) {
  startl += i;
  if (startl >= nlines) startl = max (0, nlines - 1);
}

void console::last () {
  // ensures last line is always displayed
  end ();
  if (!rollup) up (lines_per_screen - 1);
}

void console::pgdn () {
  down (lines_per_screen);
}

void console::pgup () {
  up (lines_per_screen);
}

void console::home () {
  startl = 0;
}

void console::end () {
  startl = max (0, nlines - 1);
}

console::console () {
  rollup = 0;
  command_mode = 0;
	curs_loc = 0;
	curs_locx = 0;
  hid = -1;
	ps1 = "->";
  clear ();
}

console& console::operator<< (const decoration& d) {
	dec = d;
  return *this;
}

console& console::operator<< (const string& s) {
	scratch_line.text += s;
	return *this;
}

console& console::operator<< (unsigned int i) {
  sprintf (buf, "%x", i);
  *this << buf;
  return *this;
}

console& console::operator<< (int i) {
  sprintf (buf, "%d", i);
	*this << buf;
  return *this;
}

console& console::operator<< (unsigned long  i) {
  sprintf (buf, "%lu", i);
  *this << buf;
	return *this;
}

console& console::operator<< (unsigned long long i) {
  sprintf (buf, "%llux", i);
	*this << buf;
  return *this;
}

console& console::operator<< (float f) {
  sprintf (buf, "%f", f);
  *this << buf;
  return *this;
}

console& console::operator<< (double d) {
  sprintf (buf, "%f", d);
  *this << buf;
  return *this;
}

console& console::operator<< (char c) {
	if (c == eol) {
		scratch_line.dec = dec;
		push_back (scratch_line);
		scratch_line = mesg ();
	} else scratch_line.text += c;
  return *this;
}

void console::del () {
	int len = cmd_line.text.length ();
	if (curs_loc > 0 && curs_loc <= len) {
		cmd_line.text.erase (cmd_line.text.begin() + curs_loc - 1);
		curs_locx = (int) x_after_string (cmd_line.text.substr (0, --curs_loc));
	}
}

void console::calc_draw_params () {
	line_height = get_line_height ();
	char_width = fnt.max_char_width;
  lines_per_screen = (win.top - win.bottom) / line_height - 2;
	startx = win.left + ps1.length () * char_width;
	starty = win.top - line_height;
}

void console::draw () {

  glMatrixMode (GL_MODELVIEW);

  glPushMatrix ();
  glLoadIdentity ();

  glMatrixMode (GL_PROJECTION);
  glPushMatrix ();
  glLoadIdentity ();
  glOrtho (win.left, win.right, win.bottom, win.top, -1, 1);

  // draw lines
  int cx = startx, cy = starty;

  int n = 0, i = startl;
  while ((n++ < lines_per_screen) && (i < nlines)) {
    const mesg& linei = lines[i++];
    const decoration& dec = linei.dec;
    glColor3f (dec.r, dec.g, dec.b);
    draw_string (linei.text, cx, cy);
    cy -= line_height;
  }

  // draw last line
	glColor3f (1, 1, 1);
	draw_string (ps1, win.left, cy);
  draw_string (cmd_line.text, cx, cy, 0);

	// draw cursor
  if (command_mode) {
  	glColor3f (1, 0, 0);
  	int px = cx + curs_locx, py = cy - 2;
  	glBegin (GL_LINES);
			glVertex2i (px, py);
			glVertex2i (px + char_width, py);
		glEnd ();
	}


	glEnd ();
  glPopMatrix ();

  glMatrixMode (GL_MODELVIEW);
  glPopMatrix ();


}

void console::clear () {
  lines.clear ();
  cmd_line = mesg();
  curs_loc = 0;
  curs_locx = 0;
  nlines = 0;
  startl = 0;
}

void console::set_window (const box<int>& w) {
  win = w;
	calc_draw_params ();
}

bool console::handle_input () {

  if (command_mode) {

  	const string& norm_char = charset[0].c_str();
  	const string& shift_char = charset[1].c_str();
		static const int j = norm_char.length ();

  	for (int i = 0; i < j; ++i) {
  		if (keypressedd (norm_char[i])) {
  			if (keydown (SDLK_LSHIFT) || keydown (SDLK_RSHIFT) || keydown (SDLK_CAPSLOCK)) {
					cmd_line.text.insert (cmd_line.text.begin()+curs_loc, shift_char[i]);
				} else {
					cmd_line.text.insert (cmd_line.text.begin()+curs_loc, norm_char[i]);
				}
				++curs_loc;
				curs_locx = x_after_string (cmd_line.text.substr (0, curs_loc));
  			break;
  		}
		}

		if (keypressed (SDLK_RETURN)) {

			push_back (cmd_line);

			history.push_back (cmd_line.text);
			hid = history.size () - 1;

			operator() (cmd_line.text);
			cmd_line = mesg ();
			curs_loc = curs_locx = 0;

		} else if (keypressedd (SDLK_LEFT)) {
			if (--curs_loc < 1) {
				curs_loc = 0;
				curs_locx = 0;
			} else curs_locx = (int) x_after_string (cmd_line.text.substr (0, curs_loc));
		} else if (keypressedd (SDLK_RIGHT)) {
			int len = cmd_line.text.length ();
			if (++curs_loc > len) curs_loc = len;
			curs_locx = (int) x_after_string (cmd_line.text.substr (0, curs_loc));
		}

		if (hid > -1) {

			if (keypressedd (SDLK_UP)) {
				cmd_line.text = history [hid--];
				curs_loc = cmd_line.text.length ();
				curs_locx = x_after_string (cmd_line.text);
				if (hid < 0) hid = history.size () - 1;
			}

			if (keypressedd (SDLK_DOWN)) {
				cmd_line.text = history [hid++];
				curs_loc = cmd_line.text.length ();
				curs_locx = x_after_string (cmd_line.text);
				if (hid >= (int) history.size()) hid = 0;
			}

		}

	}

  static double sc0 = 0.5, sc1 = 1./128;
  static double pg0 = 0.5, pg1 = 1./32;
	if (keypressed (SDLK_TAB)) {
		command_mode = !command_mode;
		last ();
	} else
  if (keypressedd (SDLK_INSERT, sc0, sc1)) up(1); else
  if (keypressedd (SDLK_DELETE, sc0, sc1)) down(1); else
  if (keypressed (SDLK_HOME)) home (); else
  if (keypressed (SDLK_END)) end(); else
  if (keypressedd (SDLK_PAGEUP, pg0, pg1)) pgup(); else
  if (keypressedd (SDLK_PAGEDOWN, pg0, pg1)) pgdn(); else
  if (keypressedd (SDLK_BACKSPACE)) {
		if (command_mode) del (); else clear();
  } else
  if (keypressed ('`') && !command_mode) {
		rollup = !rollup;
		last ();
	}
  return true;

}

console& console::operator() (const string& cmd) {
	*this << console::cyan;
	extern tcl_interp interpreter;
	into_lines (interpreter(cmd), *this);
	return *this;
}

