#include "din.h"
#include "console.h"
#include "solver.h"
#include "utils.h"
#include "input.h"
#include "color.h"
#include "random.h"
#include "vector2d.h"
#include "sstream"
#include "command.h"
#include "delay.h"
#include "chrono.h"
#include "osc.h"
#include "delay.h"
#include "bot.h"

using namespace std;

extern string dotdin; // ~/.din directory
extern console cons; // console
extern viewport view; // viewport
extern int mousex, mousey; // mouse pos
extern int lmb, rmb, mmb; // mouse button state
extern int LEFT, BOTTOM, RIGHT, TOP; // din board extents
extern int HEIGHT; // TOP - BOTTOM
extern float DELTA_VOLUME; // delta volume per unit key height
extern int LAST_VOLUME; // last volume in key height units
extern int NUM_VOLUMES; // number of available volumes
extern int NUM_MICROTONES; // default number of microtones in a range
extern map<string, int> NOTE_POS; // interval name -> value, 1 - 1, 1# - 2, 2 - 3  etc
extern int SAMPLE_RATE; // sampling rate
extern map<string, float> INTERVALS; // interval name -> value
extern audio_out aout; // audio output

din::din (cmdlist& cl) :
  win (0, 0, view.xmax, view.ymax),
	drone_master_volume (0.1),
  drone_wave ("drone.crv"),
  droneed ("drone.ed"),
  dronelis (this),
	fm ("fm.pat"),
	am ("am.pat"),
	gatl ("gate-l.pat"),
	gatr ("gate-r.pat"),
	osh (this),
	is_octave_shift (0),
	octave_shift ("octave_shift.crv"),
	gated ("gater.ed"),
	gatlib ("gater_patterns.lib"),
	gatllis (&gatl),
	gatrlis (&gatr),
	octed ("octave_shift.ed"),
	octlib ("octave_shift.lib"),
	octlis (&octave_shift),
	moded ("mod.ed"),
	fmlis (&fm),
	amlis (&am),
  wavlib ("waveform.lib"),
  stred (this, "strength.ed", "strength_editor.hlp"),
  waved ("wave.ed"),
  chaned ("channels.ed"),
  strlis (this), wavelis (this),
  leftlis (this), rightlis (this),
	comed ("compressor.ed"),
	coml ("coml.crv"),
	comr ("comr.crv"),
  helptext ("din.hlp"),
  notation ("numeric")
{

		name = "din board";

		prev_mousex = 0;
		prev_mousey = view.ymax;
		win_mousex = win_mousey = 0;

    current_range = 0;

		snap_drones = 0;
		drone_sol (&drone_wave);
		droneed.add (&drone_wave, &dronelis);
		droneed.attach_library (&wavlib);

    num_resonators = 0;
    cur_resonator = -1;

    waved.attach_library (&wavlib);
		gated.attach_library (&gatlib);

		gated.add (&gatl.crv, &gatllis);
		gated.add (&gatr.crv, &gatrlis);
		gated.bv.push_back (&gatl);
		gated.bv.push_back (&gatr);

		octave_shift.xmax = &osh;
		octed.add (&octave_shift.crv, &octlis);
		octed.attach_library (&octlib);

		am_depth = 0;
		fm_depth = 0;

		moded.add (&fm.crv, &fmlis);
		moded.add (&am.crv, &amlis);
		moded.bv.push_back (&fm);
		moded.bv.push_back (&am);

		comed.add (&coml.crv, &coml.lis);
		comed.add (&comr.crv, &comr.lis);

		load_resonators (dinfo.resonators_fname);

		cons << console::green << "Press F1 for help." << eol;
		cons << console::green << "Press TAB to enter command mode." << eol;

}

void din::load_scale () {

  setup_ranges ();
  load_drones ();
	update_drone_tone ();
	update_drone_ipos (0, num_ranges);
	sync_drones ();
}

bool din::load_ranges () {
	string fname = dotdin + scaleinfo.name + ".ranges";
	ifstream file (fname.c_str(), ios::in);
	if (!file) {
		cout << "couldnt load range pos from " << fname << ", will use defaults " << eol;
		return false;
	}

	string ignore;
	file >> ignore >> ignore;

	int l = 0, r, w;
	for (int i = 0; i < num_ranges; ++i) {
		range& R = ranges[i];
		file >> ignore >> w;
		r = l + w;
		R.extents (l, BOTTOM, r, TOP);
		l = r;
	}

	return true;

}

void din::save_ranges () {

	string fname = dotdin + scaleinfo.name + ".ranges";
	ofstream file (fname.c_str(), ios::out);
	if (file) {
		cout << "<<< saving note ranges in: " << fname;
		file << "num_ranges " << num_ranges << endl;
		for (int i = 0; i < num_ranges; ++i) {
			range& r = ranges[i];
			file << i << ' ' << r.extents.width << endl;
		}
		cout << ", done >>>" << endl;
	}

}

void din::set_range_size (int ran, int sz) {

	if (ran == -1) {
		int r = LEFT, l;
		for (int i = 0; i < num_ranges; ++i) {
			l = r;
			r = l + sz;
			range& R = ranges[i];
			R.extents (l, BOTTOM, r, TOP);
		}
	} else {
		range& R = ranges[ran];
		int delta = sz - R.extents.width;
		R.extents (R.extents.left, BOTTOM, R.extents.left + sz, TOP);
		for (int i = ran + 1; i < num_ranges; ++i) {
			range& Ri = ranges[i];
			Ri.extents (Ri.extents.left + delta, Ri.extents.bottom, Ri.extents.right + delta, Ri.extents.top);
		}
	}

	update_drone_ipos (0, num_ranges);


}

void din::setup_ranges () {

	cout << "<<< setting up ranges >>>" << endl;

  num_ranges = scaleinfo.num_octaves * scaleinfo.num_ranges;
  firstr = &ranges [0];
  last_range = num_ranges - 1;
  lastr = &ranges [last_range];

	setup_range_notes ();
	notate_ranges ();

	if (load_ranges () == false) set_range_size (-1, NUM_MICROTONES);

	update_range_notes ();

	cout << "<<< ranges setup complete. >>>" << endl;

}

void din::setup_range_notes () {
  for (int p = 0, r = 0; p < scaleinfo.num_octaves; ++p) {
    for (int i = 0, j = 1; i < scaleinfo.num_ranges; ++i, ++j) {
      range& R = ranges[r++];
      R.intervals[0] = scaleinfo.notes[i];
      R.intervals[1] = scaleinfo.notes[j];
    }
  }
}

void din::update_range_notes () {
  float octave_start = scaleinfo.lo_tonic;
  for (int p = 0, r = 0; p < scaleinfo.num_octaves; ++p) {
    for (int i = 0; i < scaleinfo.num_ranges; ++i) {
			ranges[r++].calc_notes (p, octave_start, scaleinfo.intervals);
    }
    octave_start *= 2;
  }

}

void din::range_left_changed () {

  range& R = ranges [current_range];
  int old_left = R.extents.left;

	if (delta_mousex != 0) {
		R.extents (R.extents.left + delta_mousex, R.extents.bottom, R.extents.right, R.extents.top);
		cons << "num_tones = " << R.extents.width << eol;
		int delta_left = R.extents.left - old_left;
		for (int i = 0; i < current_range; ++i) {
			range& ir = ranges [i];
			ir.extents (ir.extents.left + delta_left, ir.extents.bottom, ir.extents.right + delta_left, ir.extents.top);
		}

		update_drone_ipos (0, current_range);
	}

}

void din::range_right_changed () {

  range& R = ranges [current_range];
  int old_right = R.extents.right;

	if (delta_mousex != 0) {
		R.extents (R.extents.left, R.extents.bottom, R.extents.right + delta_mousex, R.extents.top);
		cons << "num_tones = " << R.extents.width << eol;
		int delta_right = R.extents.right - old_right;
		for (int i = current_range + 1; i < num_ranges; ++i) {
			range& ir = ranges [i];
			ir.extents (ir.extents.left + delta_right, ir.extents.bottom, ir.extents.right + delta_right, ir.extents.top);
		}
		update_drone_ipos (current_range, num_ranges);
	}

}


void din::set_key_id (int i) {
	key_id = i;
	notate_ranges ();
}

void din::set_notation (const string& s) {
	if (s == "w" || s == "west" || s == "western") notation = "western";	else notation = "numeric";
	notate_ranges ();
}

void din::notate_ranges () {

	static const char* western [] = {"C", "Db", "D", "Eb", "E", "F", "F#", "G", "Ab", "A", "Bb", "B"};
	if (notation == "western") {
		for (int i = 0; i < num_ranges; ++i) {
			range& ri = ranges [i];
			string i0 = ri.intervals[0], i1 = ri.intervals[1];
			int ii0 = NOTE_POS[i0], ii1 = NOTE_POS[i1];
			int kii0 = (key_id + ii0) % 12;
			int kii1 = (key_id + ii1) % 12;
			ri.labels[0] = western [kii0];
			ri.labels[1] = western [kii1];
		}
	} else {
		for (int i = 0; i < num_ranges; ++i) {
			range& ri = ranges [i];
			string i0 = ri.intervals[0], i1 = ri.intervals[1];
			ri.labels[0] = i0;
			ri.labels[1] = i1;
		}
	}

}

void din::load_resonators (const string& fname) {

	if (fname == "") return;
	dinfo.resonators_fname = fname;

	string fresonators = dotdin + dinfo.resonators_fname;
  ifstream file (fresonators.c_str(), ios::in);
  if (!file) {
    cout << "bad resonators file: " << fname << endl;
    exit (1);
    return;
  } else {

		stred.clear ();
    string ignore;

    file >> ignore >> num_resonators;
    file >> ignore >> cur_resonator; if (cur_resonator >= num_resonators) cur_resonator = 0;

		resonators.clear ();
    for (int i = 0; i < num_resonators; ++i) {

      resonators.push_back (resonator());

			resonator& ri = resonators[i];
			file >> ignore >> ignore;
			file >> ignore >> ri.disabled;
      file >> ignore; ri.strength.load (file);
      file >> ignore; ri.wave.load (file);
      file >> ignore; ri.left.load (file);
      file >> ignore; ri.right.load (file);

      players.push_back (play());

    }

    for (int i = 0; i < num_resonators; ++i) {

			resonator& ri = resonators[i];
			ri.get_strength (&ri.strength);
			ri.get_wave (&ri.wave);

      stred.add (&ri.strength, &strlis, ri.disabled);

			render_left_volume (i);
      render_right_volume (i);

			play& pi = players[i];
			pi.set_wave (&ri.get_wave);

		}

    stred.pick (cur_resonator);
    cout << ", done. >>>" << endl;

   }

	calc_multicore_vars ();

}

void din::save_resonators (int deldis) {
	string fresonators = dotdin + dinfo.resonators_fname;
  ofstream file (fresonators.c_str(), ios::out);
  if (!file) {
    cout << "cant save resonators" << eol;
  } else {
		cout << "<<< saving resonators in: " << fresonators;
		int nr = num_resonators;
		if (deldis) for (int i = 0; i < num_resonators; ++i) if (resonators[i].disabled) --nr;
    file << "num_resonators " << nr << eol;
		clip<int> (0, cur_resonator, nr-1); file << "cur_resonator " << cur_resonator << eol;
    for (int i = 0; i < num_resonators; ++i) {
      resonator& ri = resonators[i];
      int disabled = ri.disabled;
			if (deldis && disabled); else {
				file << "resonator: " << i << eol;
				file << "disabled: " << disabled << eol;
				file << "strength:" << eol;
				ri.strength.save (file);
				file << "wave:" << eol;
				ri.wave.save (file);
				file << "left_channel:" << eol;
				ri.left.save (file);
				file << "right_channel:" << eol;
				ri.right.save (file);
			}
    }
    cout << ", done >>>" << endl;
		cons << console::yellow << "saved " << nr << " resonators to: " << dinfo.resonators_fname << eol;
  }
}

void din::set_tonic (float s, int warp) {
	if (scaleinfo.set_tonic(s)) {
		update_range_notes ();
		update_drone_tone ();
	}
}

void din::mouse2tonic () {
	// set mouse at tonic
	range& r = ranges[scaleinfo.num_ranges]; // range of middle tonic
	int wx = r.extents.left, mx = -1;
	if (wx < win.left) {
		win.left = wx;
		win.right = win.left + view.xmax;
		mx = 0;
	} else if (wx > win.right) {
		win.right = wx;
		win.left = win.right - view.xmax;
		mx = view.xmax;
	} else mx = wx - win.left;

	SDL_WarpMouse (mx, mousey);
}

float din::get_tonic () { return scaleinfo.tonic;}

float din::get_note_value (const string& s) {
	return scaleinfo.intervals[s];
}

void din::retune_note (const string& nn, float v) {
	float b4 = scaleinfo.intervals[nn];
	scaleinfo.intervals[nn] = v;
	update_range_notes ();
	update_drone_tone ();
	cons << console::green << "retuned note: " << nn << " from: " << b4 << " to " << v << eol;
}

void din::retune_note () {

	// find nearest note
	range& r = ranges[current_range];
	int left = r.extents.left, right = r.extents.right;
	int delta_left = win_mousex - left, delta_right = right - win_mousex;
	int i = 0; if (delta_left > delta_right) i = 1;
	string label(r.intervals[i]);

	float sas [] = {scaleinfo.lo_tonic, scaleinfo.tonic, scaleinfo.hi_tonic};
	float sa = sas [(int) octave_position];
	float freq = step * SAMPLE_RATE;
	float interval = freq / sa;
	label = r.intervals[i];
	float b4 = scaleinfo.intervals[label];
	scaleinfo.intervals [label] = interval;
	update_range_notes ();
	update_drone_tone ();

	cons << console::green << "retuned " << label << " from " << b4 << " to " << interval << eol;

}

void din::restore_note () {

	// find nearest note
	range& r = ranges[current_range];
	int left = r.extents.left, right = r.extents.right;
	int delta_left = win_mousex - left, delta_right = right - win_mousex;
	int i = 0; if (delta_left > delta_right) i = 1;
	string n (r.intervals[i]);
	if (n != "S") {
		n = r.intervals[i];
		cons << console::green << "restored " << n << " from " << scaleinfo.intervals[n] << " to " << INTERVALS[n] << eol;
		scaleinfo.intervals [n] = INTERVALS [n];
		update_range_notes ();
		update_drone_tone ();
	}
}

void din::restore_all_notes () {
	scaleinfo.intervals = INTERVALS;
	update_range_notes ();
	update_drone_tone ();
}

void din::save_scale () {
	save_ranges ();
	save_drones ();
	scaleinfo.save_custom_tuning ();
}

din::~din () {
	save_scale ();
  save_resonators ();
}

void din::load_drones () {
	string fdrone = dotdin + scaleinfo.name + ".drone";
  ifstream file (fdrone.c_str(), ios::in);
	current_drone = -1;
	num_drones = 0;
	drones.clear ();
  if (!file) return;
  else {
    string ignore;
    file >> ignore >> num_drones;
		cout << "<<< loading " << num_drones << " drones from: " << fdrone;
    file >> ignore >> current_drone;
    file >> ignore >> drone_master_volume;
    for (int i = 0; i < num_drones; ++i) {
      drone di;
      file >> ignore >> di.range_num;
      file >> ignore >> di.fpos;
      file >> ignore >> di.vol;
      di.dv = (int) (di.vol * HEIGHT);
      di.player.set_wave (&drone_sol);
      drones.push_back (di);
    }
		cout << ", done. >>>" << eol;
  }
}

void din::save_drones () {

  drone_wave.save ("drone.crv");
  string drone_file = dotdin + scaleinfo.name + ".drone";
  ofstream file (drone_file.c_str(), ios::out);
  if (file) {
		cout << "<<< saving drones in: " << drone_file;
		file << "num_drones " << num_drones << eol;
		file << "current_drone " << current_drone << eol;
		file << "master_volume " << drone_master_volume << eol;
		for (int i = 0; i < num_drones; ++i) {
			drone& di = drones[i];
			file << "range_number " << di.range_num << eol;
			file << "fpos: " << di.fpos << eol;
			file << "volume " << di.vol << eol;
		}
		cout << ", done. >>>" << endl;
	}
}

void din::update_drone_tone () {

	for (int i = 0; i < num_drones; ++i) {
		drone& di = drones[i];
		range& r = ranges[di.range_num];
		di.step = (1 - di.fpos) * r.notes[0].step + di.fpos * r.notes[1].step;
		di.player.set (di.step, drone_master_volume * di.vol, drone_master_volume * di.vol);
	}

}

void din::update_drone_ipos (int s, int t) {
	for (int i = 0; i < num_drones; ++i) {
		drone& di = drones[i];
		if ((di.range_num >= s) && (di.range_num < t)) {
			range& r = ranges[di.range_num];
			di.ipos = (int)((1 - di.fpos) * r.extents.left + di.fpos * r.extents.right);
		}
	}
}

void din::sync_drones () {
	cout << "sync drones: ";
	for (int i = 0; i < num_drones; ++i) {
		cout << i << ' ';
		drone& di = drones[i];
		di.player.x = 0;
	}
	cout << endl;
}

void din::add_drone () {
	drones.push_back (drone());
	current_drone = num_drones++;
	set_drone (current_drone);
}

void din::remove_drone () {

  if (current_drone == -1) return;
	drones.erase (drones.begin() + current_drone);
	--current_drone;
	--num_drones;
	// sync_drones ();

}

void din::set_drone (int d) {

    if (d == -1) return;

    find_current_range ();
		find_volume ();

    drone& dd = drones[d];

    dd.dv = dv;
    dd.vol = dv * DELTA_VOLUME;

    dd.range_num = current_range;
    range& curr = ranges[dd.range_num];

    int delta_left = win_mousex - curr.extents.left;
    if (snap_drones) {
			int delta_right = curr.extents.right - win_mousex;
			int note_num = 0; if (delta_right < delta_left) note_num = 1;
			if (note_num) {
				dd.fpos = 1;
				dd.ipos = curr.extents.right;
			} else {
				dd.fpos = 0;
				dd.ipos = curr.extents.left;
			}
			dd.step = curr.notes[note_num].step;
		} else {
			dd.fpos = delta_left * 1. / curr.extents.width;
			float one_fpos = 1 - dd.fpos;
			dd.step = one_fpos * curr.notes[0].step + dd.fpos * curr.notes[1].step;
			dd.ipos = (int)(one_fpos * curr.extents.left + dd.fpos * curr.extents.right);
		}

    dd.player.set_wave (&drone_sol);
    dd.player.set (dd.step, drone_master_volume * dd.vol, drone_master_volume * dd.vol);

		// sync_drones ();

}

void din::pick_drone () {
  current_drone = -1;
  for (int i = 0; i < num_drones; ++i) {
    drone& di = drones[i];
    int hit2 = di.dv * di.dv;
    int d2 = distance2 (di.ipos, 0, win_mousex, 0);
    if (d2 <= hit2) {
      current_drone = i;
      break;
    }
  }
}

void din::mark_drones () {
	glLineWidth (2);
	static const float cc2 = 0.25;
  for (int i = 0; i < num_drones; ++i) {
    drone& di = drones[i];
    int x = di.ipos, y = BOTTOM, ydv = y + di.dv;
    glColor3f (cc2, cc2, cc2);
    glBegin (GL_LINE_LOOP);
      glVertex2i (x - di.dv, ydv);
      glVertex2i (x, y);
      glVertex2i (x + di.dv, ydv);
    glEnd ();
  }
	glLineWidth (1);
}

void din::update_drone_master_volume (float d) {
	drone_master_volume += d;
	if (drone_master_volume < 0) drone_master_volume = d;
	for (int i = 0, j = drones.size(); i < j; ++i) {
		drone& di = drones[i];
		di.player.set_volume (drone_master_volume * di.vol, drone_master_volume * di.vol);
	}
	cons << "drone master volume = " << drone_master_volume << eol;

}

void din::calc_win_mouse () {

	delta_mousex = mousex - prev_mousex;
	delta_mousey = mousey - prev_mousey;

	win_mousex += delta_mousex;
	win_mousey -= delta_mousey;

	prev_mousex = mousex;
	prev_mousey = mousey;

}

bool din::handle_input () {

	static const float delta_drone_master = 0.01;
	static float last_drone_master_volume = 0;
	static int mod_dh = 0; // mod height

  osc.handle_input (); // oscilloscope

	if (phrasor0.state == phrasor::recording) { // record mouse pos for playback l8r
		static point<int> pt;
		pt.x = win_mousex; pt.y = win_mousey;
		phrasor0.data.push_back (pt);
		++phrasor0.sz;
	}

  // movement
  if (keydown (SDLK_a)) scroll (-dinfo.scroll.x, 0); else
  if (keydown (SDLK_d)) scroll (+dinfo.scroll.x, 0); else
  if (keydown (SDLK_w)) scroll (0, +dinfo.scroll.y); else
  if (keydown (SDLK_s)) scroll (0, -dinfo.scroll.y);


	if (keypressed (SDLK_z)) {
		if (!is_octave_shift) {
			osh.tonic = scaleinfo.tonic;
			osh.dir = 0;
			octave_shift.now = 0;
			is_octave_shift = 1;
			cons << console::yellow  << "modulating down" << eol;
		}
	} else if (keypressed (SDLK_x)) {
		if (!is_octave_shift) {
			osh.tonic = scaleinfo.tonic;
			osh.dir = 1;
			octave_shift.now = 0;
			is_octave_shift = 1;
			cons << console::yellow << "modulating up" << eol;
		}
	} else
  if (keypressed (SDLK_b)) {
		!gatl; if (gatl.enabled) cons << console::green << "gate-l enabled" << eol; else cons << console::red << "gate-l disabled" << eol;
		!gatr; if (gatr.enabled) cons << console::green << "gate-r enabled" << eol; else cons << console::red << "gate-r disabled" << eol;
	} else

  if (keypressed (SDLK_f)) {
    cons << console::yellow;
    if (phrasor0.state == phrasor::recording) {
			phrasor0.validate ();
			phrasor0.play ();
      cons << "phrasor STOPPED recording, started PLAYING." << eol;
    } else if (phrasor0.state != phrasor::playing) {
      phrasor0.rec ();
      cons << "phrasor is RECORDING." << eol;
    }
  } else

  if (keypressed (SDLK_v)) {
    if (phrasor0.state == phrasor::playing) {
      cons << console::red << "phrasor has PAUSED." << eol;
      phrasor0.state = phrasor::paused;
      find_current_range ();
    } else {
      cons << console::green << "phrasor is PLAYING." << eol;
			phrasor0.validate ();
      phrasor0.play ();
    }
  } else

  if (keypressed (SDLK_g)) {
    phrasor0.clear ();
    find_current_range ();
    cons << console::red << "phrases deleted." << eol;
  }

  // phrase jogging
  jogged = 0;
  if (keydown (SDLK_LEFT)) {phrasor0.jog (-3); jogged = 1;} else
  if (keydown (SDLK_RIGHT)) {phrasor0.jog (+3); jogged = 1;} else
  if (keydown (SDLK_DOWN)) phrasor0.set_cue (); else
  if (keydown (SDLK_UP)) phrasor0.goto_cue (); else


  // oscilloscope
  if (keypressed (SDLK_RCTRL)) {
		osc.toggle_visible ();
	} else if (keypressed (SDLK_MENU)) osc.toggle_pause ();

	// drone
	if (keypressed (SDLK_q)) {
			add_drone ();
	}
	else
	if (keypressed (SDLK_e)) {
			pick_drone ();
			if (current_drone > -1) {
				set_drone (current_drone);
			}
	}
	else
	if (keypressed (SDLK_c)) {
			pick_drone ();
			if (current_drone > -1) {
					remove_drone ();
			}
	}
	else
	if (keypressed (SDLK_p)) {
		dinfo.compress = !dinfo.compress;
		if (dinfo.compress) cons << console::green << "compressor enabled." << eol; else cons << console::red << "compressor disabled." << eol;
	} else
	if (keypressed (SDLK_o)) {
		extern oserver srv;
		srv.toggle ();
		if (srv.enabled) cons << console::green << "server enabled" << eol; else cons << console::yellow << "server paused" << eol;
	} else
	if (keypressedd (SDLK_COMMA)) update_drone_master_volume (-delta_drone_master); else
	if (keypressedd (SDLK_PERIOD)) update_drone_master_volume (+delta_drone_master); else
	if (keypressedd (SDLK_SLASH)) {
		if (drone_master_volume == 0) update_drone_master_volume (last_drone_master_volume); else {
			last_drone_master_volume = drone_master_volume;
			update_drone_master_volume (-drone_master_volume);
		}
	}
	else
	if (keypressed (SDLK_QUOTE)) {
		dinfo.delay = !dinfo.delay;
		cons << "delay ";
		if (dinfo.delay) cons << "enabled." << eol; else cons << " disabled." << eol;
	}
	else
	if (keypressed (SDLK_SEMICOLON)) {
		set_tonic (step * SAMPLE_RATE, 1);
		cons ("key");
	}
	else
  if (keydown (SDLK_LCTRL)) {
		if (keypressed (SDLK_m)) set_range_size (-1, ranges[current_range].extents.width); else
		if (keypressed (SDLK_n)) restore_all_notes (); else
		range_left_changed ();
  }
  else
  if (keydown (SDLK_LSHIFT)) {
		if (keypressed (SDLK_m)) set_range_size (-1, NUM_MICROTONES); else
		if (keypressed (SDLK_n)) restore_note (); else
		range_right_changed ();
  }
	else
	if (keypressed (SDLK_n)) {
		retune_note ();
	}
  else
	if (keypressed (SDLK_m)) {
		set_range_size (current_range, NUM_MICROTONES);
	}
	else
	if (keypressed (SDLK_i)) {
		extern oserver srv;
		srv.msg = !srv.msg;
		if (srv.msg) cons << console::green << "mesgs on" << eol; else cons << console::red << "mesgs off" << eol;
	}
	else
	if (keydown (SDLK_h)) {
		if (mod_dh == 0) {
			mod_dh = 1;
		} else {
			if (delta_mousey != 0) {
				extern int MAX_HEIGHT;
				HEIGHT -= delta_mousey;
				clip<int> (1, HEIGHT, MAX_HEIGHT);
				TOP = BOTTOM + HEIGHT;
				for (int i = 0; i < num_ranges; ++i) {
					ranges[i].extents.top = TOP;
					ranges[i].extents.height = HEIGHT;
				}

				calc_volume_vars (HEIGHT);

				for (int i = 0; i < num_drones; ++i) {
					drone& di = drones[i];
					di.dv = (int)(di.vol * HEIGHT);
				}
				cons << console::yellow << "board height = " << HEIGHT << eol;
			}
		}
	} else {
		if (mod_dh) {
			mod_dh = 0;
			render_volumes ();
		}
	}


	// change am/fm/gate-l/gate-r rate
	//
	static double rept = 1/32.;
	static beat2value* bv[] = {&am, &fm, &gatl, &gatr};
	static const char* bs [] = {"sb am", "sb fm", "sb gate-l", "sb gate-r"};
	static unsigned int bv_rate_dec[] = {SDLK_F5, SDLK_F7, SDLK_F9, SDLK_F11};
	static unsigned int bv_rate_inc[] = {SDLK_F6, SDLK_F8, SDLK_F10, SDLK_F12};
	static float bv_delta [] = {1, 1, 1, 1};
	static char buf [256]={0};
	int nbv = 4;

	extern bot bott;

	string s;
	for (int i = 0; i < nbv; ++i) {
		int hit = 0;
		if (keypressedd (bv_rate_dec[i], rept, rept)) {
			bv[i]->set_bpm (bv[i]->get_bpm() - bv_delta[i]);
			hit = 1;
		} else if (keypressedd (bv_rate_inc[i], rept, rept)) {
			bv[i]->set_bpm (bv[i]->get_bpm () + bv_delta[i]);
			hit = 1;
		}
		if (hit) {
			cons << bv[i]->crv.name << " bpm: " << bv[i]->bpm << eol;
			sprintf (buf, "%s %f\n", bs[i], bv[i]->bpm);
			s = buf; bott << s;
			break;
		}
	}

	// change am/fm amount
	nbv = 2;
	static float* bv_amt [] = {&am_depth, &fm_depth};
	static const char* bvs [] = {"sv amd ", "sv fmd "};
	static unsigned int bv_amt_dec [] = {SDLK_r, SDLK_y};
	static unsigned int bv_amt_inc [] = {SDLK_t, SDLK_u};
	static float bv_amt_delta [] = {0.025f, 1.f};
	for (int i = 0; i < nbv; ++i) {
		int hit = -1;
		if (keypressedd (bv_amt_dec[i], rept, rept)) {
			*bv_amt[i] -= bv_amt_delta[i];
			hit = i;
		} else
		if (keypressedd (bv_amt_inc [i], rept, rept)) {
			*bv_amt[i] += bv_amt_delta[i];
			hit = i;
		}
		if (hit != -1) {
			hz2step (fm_depth, fm_step);
			cons << bv[i]->crv.name << " depth: " << *bv_amt[i] << eol;
			sprintf (buf, "%s %f\n", bvs[i], *bv_amt[i]);
			s = buf; bott << s;
			break;
		}
	}

  if (keypressed (SDLK_F1)) helptext();

  return true;

}

void din::scroll (int dx, int dy) {

	mousex -= dx;
	prev_mousex -= dx;

	mousey += dy;
	prev_mousey += dy;

	win (win.left + dx, win.bottom + dy, win.right + dx, win.top + dy);
	SDL_WarpMouse (mousex, mousey);
}



void din::find_current_range () {

  win_mousex = win.left + mousex;
  win_mousey = win.bottom + view.ymax - mousey;

  if (win_mousex <= ranges[0].extents.left) current_range = 0; else
  if (win_mousex >= ranges[last_range].extents.right) current_range = last_range; else
  for (int i = 0; i < num_ranges; ++i) {
    range& curr = ranges[i];
    box<int>& ext = curr.extents;
    if ( (win_mousex >= ext.left) && (win_mousex <= ext.right)) {
      current_range = i;
      break;
    }
  }
}

bool din::find_tone_and_volume () {

	find_volume ();

	// current tone
	range* curr = &ranges [current_range];
	int deltax = win_mousex - curr->extents.left;
	if (deltax >= curr->extents.width) {
		++current_range;
		if (current_range == num_ranges) {
			current_range = last_range;
			curr = lastr;
		} else {
			curr = &ranges [current_range];
		}
	} else if (deltax < 0) {
		--current_range;
		if (current_range < 0) {
			curr = firstr;
			current_range = 0;
		} else {
			curr = &ranges [current_range];
		}
	}

	deltax = win_mousex - curr->extents.left;
	delta = deltax * curr->extents.width_1;
	step = curr->notes[0].step + delta * curr->delta_step; // current tone
	octave_position = curr->notes[0].octave_position + delta * curr->delta_octave_position;

  if (dv < 0) { // below keyboard, silence
    for (int i = 0; i < num_resonators; ++i) {
			play& p = players[i];
			p.set (step, 0, 0);
			p.x = 0;
			am_vol = 0;
			/*gatl.now = gatr.now = 0;
			fm.now = am.now = 0;*/
		}
		return false;
  } else {
		am_vol = dv * DELTA_VOLUME;
    for (int i = 0; i < num_resonators; ++i) {
      play& pi = players[i];
      if (resonators[i].disabled == false) {
        float next_left = volumes[dv].res[i].left;
        float next_right = volumes[dv].res[i].right;
        float next_strength = resonators[i].get_strength (octave_position);
        pi.set (step, next_left * next_strength, next_right * next_strength);
      } else {
				pi.set (step, 0, 0);
				pi.x = 0;
			}
    }
		return true;
  }

}

void din::render_left_volume (int r) {

  resonator& rr = resonators[r];
  solver gety (&rr.left);

  float x = DELTA_VOLUME;
  for (int i = 0; i < NUM_VOLUMES; ++i) {
    volumes[i].res[r].left = gety (x);
    x += DELTA_VOLUME;
  }

}

void din::render_right_volume (int r) {

  resonator& rr = resonators[r];
  solver gety (&rr.right);
  float x = DELTA_VOLUME;
  for (int i = 0; i < NUM_VOLUMES; ++i) {
    volumes[i].res[r].right = gety (x);
    x += DELTA_VOLUME;
  }

}

void din::render_volumes () {
  for (int i = 0; i < num_resonators; ++i) {
    render_left_volume (i);
    render_right_volume (i);
  }
}

void din::add_resonator () {

		/*resonator ri;

    ri.strength.load ("strength.default");
    ri.get_strength (&ri.strength);

    ri.wave.load ("waveform.default");
    ri.get_wave (&ri.wave);

    players[num_resonators].set_wave (&ri.get_wave);

    ri.left.load ("left.default");
    ri.right.load ("right.default");

    stred.add (&ri.strength, &strlis, false);
    ri.update_solvers ();

    render_left_volume (num_resonators);
    render_right_volume (num_resonators);

		calc_multicore_vars ();

    ++num_resonators;

    cons << "added resonator "<< eol;*/

}

void din::edit_resonator (int i, unsigned int what) {
  resonator& ri = resonators[i];
  switch (what) {
    case resonator::STRENGTH:
			ri.get_strength.update ();
    break;
    case resonator::WAVE:
			ri.get_wave.update ();
    break;
    case resonator::LEFT:
      render_left_volume(i);
    break;
    case resonator::RIGHT:
      render_right_volume(i);
    break;
  }
}

void din::select_resonator (int sel) {

    if (sel < 0 || sel >= num_resonators) return;

    cur_resonator = sel;
    resonator& rs = resonators[sel];
    waved.clear ();
    waved.add (&rs.wave, &wavelis);
		waved.dofft ();
    chaned.clear ();
    chaned.add (&rs.left, &leftlis);
    chaned.add (&rs.right, &rightlis);
    stred.select_curve (sel);
}

void din::browse_resonator (int i) {
	cur_resonator += i;
	if (cur_resonator > -1 && cur_resonator < num_resonators) stred.select_curve (cur_resonator);
	cons << console::cyan << "resonator " << cur_resonator << eol;
}

void din::disable_resonator (int i, bool dis) {
  if (i > -1 && i < resonator::MAX_RESONATORS) {
    resonator& ri = resonators[i];
    ri.disabled = dis;
    ri.edited |= resonator::STRENGTH;
    /*cons << "resonator " << i << " is ";
    if (dis == false) cons << console::green << " enabled." << eol; else cons << console::red << " disabled." << eol;*/
  }
}

void din::disable_all_resonators () {
	for (int i = 0; i < num_resonators; ++i) {
		disable_resonator (i);
	}
}

void din::draw () {

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  glOrtho (0, view.xmax, 0, view.ymax, -1, 1);

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();

	// draw oscilloscope
  if (osc.visible) {
    osc.draw ();
    glLoadIdentity ();
  }

	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
  glOrtho (win.left, win.right, win.bottom, win.top, -1, 1);

  // draw ranges
  int rl = firstr->extents.left, rr = lastr->extents.right;
	float tr = 1; glColor3f (tr, tr, tr);
  glBegin (GL_LINES);
    glVertex2i (rl, BOTTOM);
    glVertex2i (rr, BOTTOM);
    glVertex2i (rl, TOP);
    glVertex2i (rr, TOP);
  glEnd ();

  // label the ranges
  for (int i = 0; i < last_range; ++i) ranges[i].draw_labels ();
  lastr->draw_labels (range::BOTH);

	mark_drones ();

	// phrasor markers
	if (phrasor0.state != phrasor::stopped) {
		if (phrasor0.state == phrasor::recording) phrasor0.draw (phrasor0.recl, 1, 0, 0); else phrasor0.draw (phrasor0.cur, 0, 1, 0);
	}

	/*// draw cursor
	glBegin (GL_LINES);
		glColor3f (1, 1, 1);
		static const int sz = 10;
		glVertex2i (win_mousex - sz, win_mousey);
		glVertex2i (win_mousex + sz, win_mousey);
		glVertex2i (win_mousex, win_mousey - sz);
		glVertex2i (win_mousex, win_mousey + sz);
	glEnd ();*/

}

void din::enter () {

	if (phrasor0.state != phrasor::playing) {

		SDL_WarpMouse (prev_mousex, prev_mousey);
		mousex = prev_mousex;
		mousey = prev_mousey;

		find_current_range ();

	}

}

void din::change_depth (int i, float d) {
  if (i == 1) {
    fm_depth += d;
    hz2step (fm_depth, fm_step);
    cons << "fm depth: " << fm_depth << eol;
  } else {
    am_depth += d;
    cons << "am depth: " << am_depth << eol;
  }
}

void din::fill () {

	am.modulate (aout.ams, aout.result, aout.samples_per_channel, am_depth * am_vol);
	fm.modulate (aout.fms, aout.result, aout.samples_per_channel, fm_step);

}

void din::bg () {

  if (phrasor0.state == phrasor::playing) {
		phrasor0.get (win_mousex, win_mousey);
    if (!jogged) phrasor0.next ();
  }

  if (aout.available [aout.writei]) return;

  float* out0 = aout.writep;
	float* out1 = out0 + 1;
	int fpb = aout.samples_per_channel;

	memset (out0, 0, aout.samples_buffer_size); // silence

	// find tone/volume from mouse position, take contrib from each resonator, do fm/am
	if (find_tone_and_volume ()) {

		fill ();
		for (int i = 0; i < num_resonators; ++i) {
			float* rout = out0;
			players[i] (rout, fpb, aout.result, aout.ams, aout.fms);
		}


		// apply gaters
		//

		if (gatl.enabled) { // left gater
			float* lout = out0;
			gatl (lout, aout.result, fpb);
		}


		if (gatr.enabled) { // right gater
			float* rout = out1;
			gatr (rout, aout.result, fpb);
		}

	}

	// apply octave shift
	//
	if (is_octave_shift) {

		octave_shift (aout.result, fpb);
		static int last = fpb - 1;
		float shift_amt = aout.result [last];

		if (osh.dir) {
			set_tonic (osh.tonic * shift_amt);
		} else {
			if (shift_amt != 0) set_tonic (osh.tonic / shift_amt);
		}

		if (!is_octave_shift) {
			static const string KEY = "key";
			cons << console::yellow << "modulation complete" << eol;
			cons (KEY);
		}

	}

	// apply drones
	//

	for (int i = 0, j = num_drones; i < j; ++i) {
		float* dout = out0;
		drones[i].player (dout, fpb, aout.result);
	}

	// apply delay
	//

	extern delay left_delay, right_delay;
	if (dinfo.delay) {
		float* outl = out0; left_delay (outl, fpb);
		float* outr = out1; right_delay (outr, fpb);
	}


	// apply compression
	//
	if (dinfo.compress) {
		float* outl = out0;
		float* outr = out1;
		coml.apply (outl, fpb);
		comr.apply (outr, fpb);
	}

	// write audio to oscilloscope
	if (!osc.paused && osc.visible) osc.add_samples (out0, fpb);

	aout.available [aout.writei] = 1;
	if (++aout.writei >= aout.num_samples_buffers) {
		aout.writep = aout.samples_buffers;
		aout.writei = 0;
	} else aout.writep += aout.samples_per_buffer;

}

din_info::din_info () {

	extern string dotdin;
	ifstream file ((dotdin + "din_info").c_str(), ios::in);
	string ignore;
	if (file) {
		file >> ignore >> resonators_fname;
		file >> ignore >> scroll.x >> scroll.y;
		file >> ignore >> height; HEIGHT = height; TOP = BOTTOM + HEIGHT; calc_volume_vars (HEIGHT);
		file >> ignore >> cons.rollup;
		file >> ignore >> ncores;
		file >> ignore >> delay;
		file >> ignore >> compress;
	} else {
		resonators_fname = "one";
		scroll = point<int>(1, 1);
		height = 300;
		cons.rollup = false;
		ncores = 1;
		delay = 1;
		compress = 1;
	}
	cons.last ();
}

din_info::~din_info () {
	extern string dotdin;
	ofstream file ((dotdin + "din_info").c_str(), ios::out);
	if (file) {
		file << "resonators_file " << resonators_fname << endl;
		file << "scroll " << scroll.x << ' ' << scroll.y << endl;
		file << "board_height " << HEIGHT << endl;
		file << "rollup " << cons.rollup << endl;
		file << "ncores " << ncores << endl;
		file << "delay " << delay << endl;
		file << "compress " << compress << endl;
	} else {
		cout << "cannot save din_info" << eol;
	}
}

void din::calc_multicore_vars () {
	dinfo.resonators_per_core = (int)(num_resonators * 1./ dinfo.ncores + 1);
	dinfo.last_resonator = num_resonators - 1;
}

void octave_shift_handler::operator () (solver& s, float& x, float& dx) {
	x = s.lastx;
	d->is_octave_shift = 0;
}
