/*
 * din
 * Copyright(c) 2006-2010 S Jagannathan
 * Distributed under GNU 2.0 license.
 * Please see file COPYING for more information.
 * din is Free Software.
*/
#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>
#include <tcl.h>
#include "box.h"
#include "utils.h"
#include "font.h"
#include "console.h"
#include "input.h"
#include "viewwin.h"
#include "curve.h"
#include "multi_curve.h"
#include "solver.h"
#include "chrono.h"
#include "ui_list.h"
#include "curve_editor.h"
#include "font_editor.h"
#include "din.h"
#include "curve_library.h"
#include "console_iterator.h"
#include "delay.h"
#include "random.h"
#include "globals.h"
#include "command.h"
#include "scalelist.h"
#include "morse_code.h"
#include "fft.h"
#include "sine_mixer.h"
#include "checkdotdin.h"
#include "osc.h"
#include "bot.h"
#include "tcl_interp.h"
#include <vector>
#include <iostream>
#include <fstream>
#include <list>
#include <map>
#include <algorithm>

using namespace std;

// keyboard layout
string norm_us =    "abcdefghijklmnopqrstuvwxyz0123456789 .=-/;\\,[]'`";
string shift_us = 	"ABCDEFGHIJKLMNOPQRSTUVWXYZ)!@#$%^&*( >+_?:|<{}\"~";
string norm_uk =    "abcdefghijklmnopqrstuvwxyz0123456789 .=-/;\\,[]'`#";
string shift_uk = 	"ABCDEFGHIJKLMNOPQRSTUVWXYZ)!\"#$%^&*( >+_?:|<{}@~~";

// ~/.din contains defaults, user savings and preferences
string dotdin;
checkdotdin cdd;

// custom vector font
font fnt ("jag.fnt");

// console
console cons;

// initial globals
load_globals lg;

// display - using OpenGL
extern int SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_DEPTH;
viewport view (SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_DEPTH);

// list of scales
scalelist scalelst;

// keyboard and mouse state
keyboard keybd;
int mousex, mousey;
int lmb, rmb, mmb;

// L and R delays
delay left_delay (1000, 0.75, 0.3, "feedback-l.crv", "volume-l.crv"), right_delay (1000, 0.75, 0.3, "feedback-r.crv", "volume-r.crv");
curve_editor delayed ("delay.ed");

void setup_delay_editor () {
	delayed.add (&left_delay.fbk_crv, &left_delay.fbk_lis);
	delayed.add (&left_delay.vol_crv, &left_delay.vol_lis);
	delayed.add (&right_delay.fbk_crv, &right_delay.fbk_lis);
	delayed.add (&right_delay.vol_crv, &right_delay.vol_lis);
}

// midi bpm from external midi clock
double MIDI_BPM = 0;

// tap bpm from computer keyboard
double TAP_BPM = 0;

// din clock
chrono clk;

fft fft0; // fast fourier transform (via fftwf)
sine_mixer sinemixer; // sine mixer for waveform edit

cmdlist cmdlst; // list of din commands

// din. see din.cc/.h
din din0 (cmdlst);

// audio output
audio_out aout;

// global commands, available from all editors
//
// format: long name, short name, help or help_file_name
//

// for L and R delays
struct delay;
extern delay left_delay, right_delay;
set_delay sd (&left_delay, &right_delay, "set-delay", "sd");
get_delay gd (&left_delay, &right_delay, "get-delay", "gd");

// for adding/listing/removing scales
add_scale as ("add-scale", "as");
remove_scale rs ("remove-scale", "rs");
ls_scale ls ("ls-scale", "ls");
ls_notes lsn ("ls-notes", "lsn");
load_scale los (&din0, "load-scale", "los");

// set and get din var
setv sv (&din0, "set-var", "sv");
getv gv (&din0, "get-var", "gv");

// set/get/ls bpm
set_bpm sb ("set-bpm", "sb");
get_bpm gb ("get-bpm", "gb");
ls_bpm lb ("ls-bpm", "lb");

// set/get bpm style
set_style ss ("set-style", "ss");
get_style gs ("get-style", "gs");

// on/off gaters
gaters gt (&din0, "gaters", "gaters");

// set key/tonic
key ky (&din0, "key", "key");

// set tuning
load_tuning ltu (&din0, "load-tuning", "ltu");

// load/save resonators that make din sound
load_resonators lre (&din0, "load-resonators", "lre");
save_resonators sre (&din0, "save-resonators", "sre");
delete_disabled_resonators ddr (&din0, "delete-disabled-resonators", "ddr");

// display notation on the keyboard
notation no (&din0, "notation", "no");

// music utils
note_distance nd ("note-distance", "nd");
chord ch ("chord", "ch");

// font cmds
set_font_size sfs ("set-font-size", "sfs");
get_font_size gfs ("get-font-size", "gfs");
set_kern sk ("set-kern", "sk");
get_kern gk ("get-kern", "gk");

// curve cmds
curve_name cn ("curve-name", "cn");
curve_value cv ("curve-value", "cv");
curve__library cl ("curve-library", "cl");
set_curve_editor sced ("set-curve-editor", "sced");
get_curve_editor gced ("get-curve-editor", "gced");
paste_gaters pasg ("paste-gaters", "paste_gaters");

// morse code
morse_code mc ("morse-code", "mc");

// unix like echo
echo ech ("echo", "say");

// OSC server/client commands
lo_address client  = 0;
oserver srv ("server", "ore");
oconnect co (client, "connect", "oco");
osend snd (client, "send", "osnd");

// IRC bot
bot bott ("bot", ":");

// OSS parameters set/query
audio__out ao (din0, "aout", "ao");

// MIDI
midi mid ("midi", "midi");

set_kb_layout kbl ("set-kb-layout", "kbl");

// oscilloscope
set_scope ssco ("set-scope", "ssco", &din0);
get_scope gsco ("get-scope", "gsco", &din0);

// drone
set_drone sdro ("set-drone", "sdro", din0);
get_drone gdro ("get-drone", "gdro", din0);

// console
set_text_color stc ("set-text-color", "stc");

// tcl interpreter. all din commands are available in the interpreter.
tcl_interp interpreter;

/*struct work_t {
	int start, end, fpb;
	float buf [audio_out::SAMPLES_PER_BUFFER];
};

const int MAX_CORES = 8;
work_t work [MAX_CORES];
pthread_t tid [MAX_CORES];

void* dowork (void* arg) {

	work_t* w = (work_t*) arg;

  memset (w->buf, 0, audio_out::FBUFFER_SIZE);

	int start = w->start, end = w->end, fpb = w->fpb;
	float* buf = w->buf;
	for (int i = start; i < end; ++i) {
		float* buf1 = buf;
		din0.players[i] (buf1, fpb, din0.ams, din0.fms);
	}
	return 0;
}*/


#define DEFINE_TCL_CMD(X) inline int (X) (ClientData cd, Tcl_Interp* ti, int objc, Tcl_Obj* CONST objv[]) { return tcl_run (cd, ti, objc, objv);}

DEFINE_TCL_CMD(tcl_key)
DEFINE_TCL_CMD(tcl_setv)
DEFINE_TCL_CMD(tcl_getv)
DEFINE_TCL_CMD(tcl_set_delay)
DEFINE_TCL_CMD(tcl_get_delay)
DEFINE_TCL_CMD(tcl_set_bpm)
DEFINE_TCL_CMD(tcl_get_bpm)
DEFINE_TCL_CMD(tcl_ls_bpm);
DEFINE_TCL_CMD(tcl_set_style)
DEFINE_TCL_CMD(tcl_get_style)
DEFINE_TCL_CMD(tcl_gaters)
DEFINE_TCL_CMD(tcl_add_scale)
DEFINE_TCL_CMD(tcl_ls_scale)
DEFINE_TCL_CMD(tcl_remove_scale)
DEFINE_TCL_CMD(tcl_load_scale)
DEFINE_TCL_CMD(tcl_ls_notes)
DEFINE_TCL_CMD(tcl_set_kern)
DEFINE_TCL_CMD(tcl_get_kern)
DEFINE_TCL_CMD(tcl_set_font_size)
DEFINE_TCL_CMD(tcl_get_font_size)
DEFINE_TCL_CMD(tcl_load_resonators)
DEFINE_TCL_CMD(tcl_save_resonators)
DEFINE_TCL_CMD(tcl_delete_disabled_resonators)
DEFINE_TCL_CMD(tcl_note_distance)
DEFINE_TCL_CMD(tcl_chord)
DEFINE_TCL_CMD(tcl_notation)
DEFINE_TCL_CMD(tcl_echo)
DEFINE_TCL_CMD(tcl_curve_value)
DEFINE_TCL_CMD(tcl_curve_name)
DEFINE_TCL_CMD(tcl_curve_library)
DEFINE_TCL_CMD(tcl_morse_code)
DEFINE_TCL_CMD(tcl_bot)
DEFINE_TCL_CMD(tcl_audio_out);
DEFINE_TCL_CMD (tcl_set_editor);
DEFINE_TCL_CMD (tcl_get_editor);
DEFINE_TCL_CMD (tcl_midi);
DEFINE_TCL_CMD (tcl_kb_layout);
DEFINE_TCL_CMD (tcl_set_scope);
DEFINE_TCL_CMD (tcl_get_scope);
DEFINE_TCL_CMD (tcl_set_drone);
DEFINE_TCL_CMD (tcl_get_drone);
DEFINE_TCL_CMD (tcl_load_tuning);
DEFINE_TCL_CMD (tcl_text_color);
DEFINE_TCL_CMD (tcl_paste_gaters);

void add_commands (Tcl_Interp* interp) {

	// ++ din commands
	//
	int ncmds = 44;
	tclcmd cmd_funcs [] = {
		tcl_key, tcl_setv, tcl_getv, tcl_set_delay, tcl_get_delay, tcl_set_bpm, tcl_get_bpm, tcl_ls_bpm, tcl_set_style, tcl_get_style,
		tcl_gaters, tcl_add_scale, tcl_ls_scale, tcl_remove_scale, tcl_load_scale, tcl_ls_notes, tcl_set_kern, tcl_get_kern, tcl_set_font_size,
		tcl_get_font_size, tcl_load_resonators, tcl_save_resonators, tcl_delete_disabled_resonators, tcl_note_distance,
		tcl_chord, tcl_notation, tcl_echo, tcl_curve_value, tcl_curve_name, tcl_curve_library, tcl_morse_code, tcl_bot,
		tcl_audio_out, tcl_set_editor, tcl_get_editor, tcl_midi, tcl_kb_layout, tcl_set_scope,
		tcl_get_scope, tcl_get_drone, tcl_set_drone, tcl_load_tuning, tcl_text_color, tcl_paste_gaters

	};

	command* cmds [] = {
		&ky, &sv, &gv, &sd, &gd, &sb, &gb, &lb, &ss, &gs,
		&gt, &as, &ls, &rs, &los, &lsn, &sk, &gk, &sfs,
		&gfs, &lre, &sre, &ddr, &nd,
		&ch, &no, &ech, &cv, &cn, &cl, &mc, &bott,
		&ao, &sced, &gced, &mid, &kbl, &ssco, &gsco, &gdro, &sdro, &ltu, &stc, &pasg
	};

	for (int i = 0; i < ncmds; ++i) {
		command* cmdp = cmds[i];
		cmdlst.add (cmdp);
		Tcl_CreateObjCommand (interp, cmdp->longname.c_str(), cmd_funcs[i], (void *) i, 0);
		Tcl_CreateObjCommand (interp, cmdp->shortname.c_str(), cmd_funcs[i], (void *) i, 0);
	}

	cout << "+++ added " << ncmds << " din commands to the Tcl interpreter +++" << endl;


}


ui_list uis;

void setup_window (int w, int h) {

	w = view.width; h = view.height;

	fft0.lev.chkpos ();
	sinemixer.sine_levels.chkpos ();

	--w; --h;
	cons.set_window (box<int>(0, -h, w, 0));

	din0.win (0, 0, w, h);
	din0.find_current_range ();
	din0.find_tone_and_volume ();

	/*din0.prev_mousex = -1;
	din0.prev_mousey = -1;
	din0.enter ();*/

}

void restore_last_window () {

	extern int SCREEN_WIDTH, SCREEN_HEIGHT;
	int width = SCREEN_WIDTH, height = SCREEN_HEIGHT;
	string ignore;

	extern string dotdin;
	ifstream file ((dotdin + "window").c_str(), ios::in); // ~/.din/window
	if (file) {
		file >> ignore >> width;
		file >> ignore >> height;
	}
	setup_window (width, height);
}

void save_window () {
	cout << "+++ saving window: ";
	const char* fname = "window";
	extern string dotdin;
	ofstream file ((dotdin + fname).c_str(), ios::out);
	extern viewport view;
	if (file) {
		file << "width " << view.width << endl;
		file << "height " << view.height << endl;
		cout << "done. +++" << endl;
	} else {
		cout << " not saved!" << endl;
	}
}

bool read_input () {

	static SDL_Event event;
    while (SDL_PollEvent(&event)) {
      switch(event.type) {
        case SDL_VIDEORESIZE:
          try {
						view.setup_surface (event.resize.w, event.resize.h, SCREEN_DEPTH);
            setup_window (event.resize.w, event.resize.h);
          } catch (const badviewport& bv) {
            cout << "sdl: viewport init failed" << endl;
            return false;
          }
          break;
        case SDL_QUIT:
					extern int quit_now;
					quit_now = 1;
					break;
      }
    }

	keybd.read ();

	int buttons = SDL_GetMouseState (&mousex, &mousey);
	lmb = buttons & SDL_BUTTON_LMASK;
	rmb = buttons & SDL_BUTTON_RMASK;
	mmb = buttons & SDL_BUTTON_MMASK;

	return true;

}

int quit_now = 0;

int main (int argc, char** argv) {

  int ignore = system (". ~/.din/m00"); // xset m 0 0 for pure mouse data, without any fudge applied by X.

  seed_rand_gen (clock()); // seed random number generator

  setup_window (SCREEN_WIDTH, SCREEN_HEIGHT);

  setup_delay_editor ();

  cons.home ();

  uis.set_current (&din0, 0); // show din board

  // bring up Tcl interpreter
	string source = "source " + dotdin;

	string load_init_tcl (source + "init.tcl"); // init scripts
	interpreter (load_init_tcl);

	string load_settings_tcl (source + "settings.tcl"); // init din settings
	interpreter (load_settings_tcl);

  aout.start (); // start audio loop in a separate thread. see audio_wanted (..)

  // ui loop

  while (read_input () && uis.handle_input () && !quit_now) {
		osc (); // handle OSC messages
		mid.handle_input (); // handle MIDI messages
		bott.handle_input (); // handle mesgs/cmds from irc
    uis.draw ();
    keybd.save ();
  }

  // saves

  save_window ();
	string save_settings_tcl ("source " + dotdin + "save_settings.tcl");
	interpreter (save_settings_tcl);

  ignore = system ("xset m $ACCEL $THRESHOLD"); // restore mouse settings

	// stop audio
	aout.stop ();

  SDL_Quit ();

  return 0;
}
